import logging
import re

import httpx
from fastapi import HTTPException

from llm_bridge.client.implementations.gemini.gemini_response_handler import *
from llm_bridge.client.model_client.gemini_client import GeminiClient
from llm_bridge.type.chat_response import ChatResponse


class NonStreamGeminiClient(GeminiClient):
    async def generate_non_stream_response(self) -> ChatResponse:
        try:
            logging.info(f"messages: {self.messages}")

            response = await self.client.aio.models.generate_content(
                model=self.model,
                contents=self.messages,
                config=self.config,
            )

            gemini_response_handler = GeminiResponseHandler()
            return await gemini_response_handler.process_gemini_response(response)

        except httpx.HTTPStatusError as e:
            status_code = e.response.status_code
            text = e.response.text
            raise HTTPException(status_code=status_code, detail=text)
        except Exception as e:
            logging.exception(e)
            match = re.search(r'\d{3}', str(e))
            if match:
                error_code = int(match.group(0))
            else:
                error_code = 500

            raise HTTPException(status_code=error_code, detail=str(e))



import logging
import re

import httpx
from anthropic import AsyncAnthropic
from anthropic.types.beta import BetaMessage
from fastapi import HTTPException

from llm_bridge.client.implementations.claude.claude_token_counter import count_claude_output_tokens
from llm_bridge.client.model_client.claude_client import ClaudeClient
from llm_bridge.type.chat_response import ChatResponse
from llm_bridge.type.serializer import serialize


async def process_claude_non_stream_response(
        message: BetaMessage,
        input_tokens: int,
        client: AsyncAnthropic,
        model: str,
) -> ChatResponse:
    text = ""
    thought = ""

    for content in message.content:
        if content.type == "thinking":
            thought += content.thinking
        if content.type == "text":
            text += content.text

    chat_response = ChatResponse(
        text=text,
        thought=thought,
    )
    output_tokens = await count_claude_output_tokens(
        client=client,
        model=model,
        chat_response=chat_response,
    )
    return ChatResponse(
        text=text,
        thought=thought,
        input_tokens=input_tokens,
        output_tokens=output_tokens,
    )


class NonStreamClaudeClient(ClaudeClient):
    async def generate_non_stream_response(self) -> ChatResponse:
        try:
            logging.info(f"messages: {self.messages}")

            message = await self.client.beta.messages.create(
                model=self.model,
                max_tokens=self.max_tokens,
                temperature=self.temperature,
                system=self.system,
                messages=serialize(self.messages),
                thinking=self.thinking,
                betas=self.betas,
                tools=self.tools,
            )

            return await process_claude_non_stream_response(
                message=message,
                input_tokens=self.input_tokens,
                client=self.client,
                model=self.model,
            )
        except httpx.HTTPStatusError as e:
            status_code = e.response.status_code
            text = e.response.text
            raise HTTPException(status_code=status_code, detail=text)
        except Exception as e:
            logging.exception(e)
            match = re.search(r'\d{3}', str(e))
            if match:
                error_code = int(match.group(0))
            else:
                error_code = 500

            raise HTTPException(status_code=error_code, detail=str(e))



from anthropic import BetaMessageStreamEvent, AsyncAnthropic

from llm_bridge.client.implementations.claude.claude_token_counter import count_claude_output_tokens
from llm_bridge.type.chat_response import ChatResponse


class ClaudeStreamResponseHandler:
    async def process_claude_stream_response(
            self,
            event: BetaMessageStreamEvent,
            input_tokens: int,
            client: AsyncAnthropic,
            model: str,
    ) -> ChatResponse:
        text = ""
        thought = ""

        if event.type == "content_block_delta":
            if event.delta.type == "thinking_delta":
                thought += event.delta.thinking
            elif event.delta.type == "text_delta":
                text += event.delta.text

        chat_response = ChatResponse(
            text=text,
            thought=thought,
        )
        output_tokens = await count_claude_output_tokens(
            client=client,
            model=model,
            chat_response=chat_response,
        )
        return ChatResponse(
            text=text,
            thought=thought,
            input_tokens=input_tokens,
            output_tokens=output_tokens,
        )



from pprint import pprint

import anthropic
from anthropic.types import TextBlockParam

from llm_bridge.type.chat_response import ChatResponse
from llm_bridge.type.model_message.claude_message import ClaudeMessage, ClaudeRole
from llm_bridge.type.serializer import serialize


async def count_claude_input_tokens(
        client: anthropic.AsyncAnthropic,
        model: str,
        system: str,
        messages: list[ClaudeMessage],
) -> int:
    response = await client.messages.count_tokens(
        model=model,
        system=system,
        messages=serialize(messages),
    )

    return response.input_tokens


async def count_claude_output_tokens(
        client: anthropic.AsyncAnthropic,
        model: str,
        chat_response: ChatResponse,
) -> int:
    text = chat_response.text.strip()
    if text == "":
        return 0

    messages = [
        ClaudeMessage(
            role=ClaudeRole.Assistant,
            content=[
                TextBlockParam(type="text", text=text),
            ]
        ),
    ]

    response = await client.messages.count_tokens(
        model=model,
        messages=serialize(messages),
    )

    return response.input_tokens



from .logic.chat_generate.chat_client_factory import create_chat_client
from .logic.chat_generate.chat_message_converter import *
from .logic.message_preprocess.message_preprocessor import preprocess_messages
from .logic.model_prices import ModelPrice, get_model_prices, find_model_prices, calculate_chat_cost
from .type.chat_response import Citation, ChatResponse
from .type.message import Role, Message, Content, ContentType
from .type.serializer import serialize



import base64
import mimetypes
from typing import Optional

from google.genai import types

from llm_bridge.client.implementations.gemini.gemini_token_counter import count_gemini_tokens
from llm_bridge.client.implementations.printing_status import PrintingStatus
from llm_bridge.type.chat_response import Citation, ChatResponse, File


class GeminiResponseHandler:
    def __init__(self):
        self.printing_status: Optional[PrintingStatus] = None
        self.prev_output_tokens: int = 0
        self.prev_printing_status: Optional[PrintingStatus] = None

    async def process_gemini_response(
            self,
            response: types.GenerateContentResponse,
    ) -> ChatResponse:
        text: str = ""
        thought: str = ""
        code: str = ""
        code_output: str = ""
        files: list[File] = []
        display: Optional[str] = None
        citations: list[Citation] = extract_citations(response)
        input_tokens, stage_output_tokens = await count_gemini_tokens(response)

        printing_status = None
        if candidates := response.candidates:
            if candidates[0].content.parts:
                for part in response.candidates[0].content.parts:
                    if part.text is not None:
                        # Thought
                        if part.thought:
                            printing_status = PrintingStatus.Thought
                            thought += part.text
                        # Text
                        elif not part.thought:
                            printing_status = PrintingStatus.Response
                            text += part.text
                    # Code
                    if part.executable_code is not None:
                        code += part.executable_code.code
                    # Code Output
                    if part.code_execution_result is not None:
                        code_output += part.code_execution_result.output
                    # File
                    if part.inline_data is not None:
                        mime_type = part.inline_data.mime_type
                        extension = mimetypes.guess_extension(mime_type) or ""
                        file = File(
                            name=f"generated_file{extension}",
                            data=base64.b64encode(part.inline_data.data).decode('utf-8'),
                            type=mime_type,
                        )
                        files.append(file)

        # Grounding Sources
        if candidates := response.candidates:
            if grounding_metadata := candidates[0].grounding_metadata:
                if search_entry_point := grounding_metadata.search_entry_point:
                    display = search_entry_point.rendered_content
                if grounding_metadata.grounding_chunks:
                    text += "\n\n# Grounding Sources:\n"
                    for i, chunk in enumerate(grounding_metadata.grounding_chunks, start=1):
                        if chunk.web:
                            text += f"{i}. [{chunk.web.title}]({chunk.web.uri})\n"

        if printing_status == self.prev_printing_status and printing_status == PrintingStatus.Response:
            output_tokens = stage_output_tokens - self.prev_output_tokens
        else:
            output_tokens = stage_output_tokens

        self.prev_output_tokens = stage_output_tokens
        self.prev_printing_status = printing_status

        return ChatResponse(
            text=text,
            thought=thought,
            code=code,
            code_output=code_output,
            files=files,
            display=display,
            citations=citations,
            input_tokens=input_tokens,
            output_tokens=output_tokens,
        )


def extract_citations(response: types.GenerateContentResponse) -> list[Citation]:
    citations = []
    if candidates := response.candidates:
        if grounding_metadata := candidates[0].grounding_metadata:
            if grounding_supports := grounding_metadata.grounding_supports:
                for grounding_support in grounding_supports:
                    citation_indices = [index + 1 for index in grounding_support.grounding_chunk_indices]
                    citation_text = grounding_support.segment.text
                    citations.append(Citation(text=citation_text, indices=citation_indices))
    return citations



import logging
import re
from pprint import pprint
from typing import AsyncGenerator, Optional

import httpx
import openai
from fastapi import HTTPException
from openai import APIStatusError, AsyncStream
from openai.types.responses import ResponseStreamEvent

from llm_bridge.client.implementations.openai.openai_token_couter import count_openai_responses_input_tokens, \
    count_openai_output_tokens
from llm_bridge.client.model_client.openai_client import OpenAIClient
from llm_bridge.type.chat_response import ChatResponse, Citation, File
from llm_bridge.type.serializer import serialize


def process_delta(event: ResponseStreamEvent) -> ChatResponse:
    text: str = ""
    files: list[File] = []
    citations: list[Citation] = []

    if event.type == "response.output_text.delta":
        text = event.delta
    # Citation is unavailable in OpenAI Responses API
    if event.type == "response.output_text.annotation.added":
        pass
    # Image Generation untestable due to organization verification requirement
    # if event.type == "response.image_generation_call.partial_image":
    #     file = File(
    #         name="generated_image.png",
    #         data=event.partial_image_b64,
    #         type="image/png",
    #     )
    #     files.append(file)

    chat_response = ChatResponse(
        text=text,
        files=files,
        citations=citations,
    )
    return chat_response


async def generate_chunk(
        stream: AsyncStream[ResponseStreamEvent],
        input_tokens: int,
) -> AsyncGenerator[ChatResponse, None]:
    try:
        async for event in stream:
            chat_response = process_delta(event)
            output_tokens = count_openai_output_tokens(chat_response)
            yield ChatResponse(
                text=chat_response.text,
                files=chat_response.files,
                citations=chat_response.citations,
                input_tokens=input_tokens,
                output_tokens=output_tokens,
            )
    except Exception as e:
        logging.exception(e)
        yield ChatResponse(error=repr(e))


class StreamOpenAIResponsesClient(OpenAIClient):
    async def generate_stream_response(self) -> AsyncGenerator[ChatResponse, None]:
        try:
            logging.info(f"messages: {self.messages}")

            input_tokens = count_openai_responses_input_tokens(
                messages=self.messages
            )

            stream: AsyncStream[ResponseStreamEvent] = await self.client.responses.create(
                model=self.model,
                reasoning=self.reasoning,
                input=serialize(self.messages),
                temperature=self.temperature,
                stream=True,
                tools=self.tools,
            )

        except httpx.HTTPStatusError as e:
            status_code = e.response.status_code
            text = e.response.text
            raise HTTPException(status_code=status_code, detail=text)
        except openai.BadRequestError as e:
            status_code = e.status_code
            text = e.message
            raise HTTPException(status_code=status_code, detail=text)
        except APIStatusError as e:
            status_code = e.status_code
            text = e.message
            raise HTTPException(status_code=status_code, detail=text)
        except Exception as e:
            logging.exception(e)
            match = re.search(r'\d{3}', str(e))
            if match:
                error_code = int(match.group(0))
            else:
                error_code = 500

            raise HTTPException(status_code=error_code, detail=str(e))

        async for chunk in generate_chunk(stream, input_tokens):
            yield chunk


import re

import openai
from fastapi import HTTPException
from openai.types import Reasoning
from openai.types.responses import WebSearchToolParam
from openai.types.responses.tool_param import CodeInterpreter, CodeInterpreterContainerCodeInterpreterToolAuto, \
    ImageGeneration

from llm_bridge.client.implementations.openai.non_stream_openai_client import NonStreamOpenAIClient
from llm_bridge.client.implementations.openai.non_stream_openai_responses_client import NonStreamOpenAIResponsesClient
from llm_bridge.client.implementations.openai.steam_openai_responses_client import StreamOpenAIResponsesClient
from llm_bridge.client.implementations.openai.stream_openai_client import StreamOpenAIClient
from llm_bridge.logic.chat_generate.chat_message_converter import convert_messages_to_openai_responses, \
    convert_messages_to_openai
from llm_bridge.type.message import Message


async def create_openai_client(
        messages: list[Message],
        model: str,
        api_type: str,
        temperature: float,
        stream: bool,
        api_keys: dict
):
    if api_type == "OpenAI":
        client = openai.AsyncOpenAI(
            api_key=api_keys["OPENAI_API_KEY"],
        )
    elif api_type == "OpenAI-Azure":
        client = openai.AsyncAzureOpenAI(
            api_version="2025-03-01-preview",
            azure_endpoint=api_keys["AZURE_API_BASE"],
            api_key=api_keys["AZURE_API_KEY"],
        )
    elif api_type == "OpenAI-GitHub":
        client = openai.AsyncOpenAI(
            base_url="https://models.inference.ai.azure.com",
            api_key=api_keys["GITHUB_API_KEY"],
        )
    elif api_type == "Grok":
        client = openai.AsyncOpenAI(
            base_url="https://api.x.ai/v1",
            api_key=api_keys["XAI_API_KEY"],
        )
    else:
        raise HTTPException(status_code=500, detail="API Type not matched")

    if api_type in ("OpenAI", "OpenAI-Azure", "Grok"):
        use_responses_api = True
    else:
        use_responses_api = False

    if use_responses_api:
        openai_messages = await convert_messages_to_openai_responses(messages)
    else:
        openai_messages = await convert_messages_to_openai(messages)

    tools = []
    reasoning = None

    if model not in ["gpt-5-chat-latest", "gpt-5-pro"]:
        tools.append(
            CodeInterpreter(
                type="code_interpreter",
                container=CodeInterpreterContainerCodeInterpreterToolAuto(type="auto")
            )
        )
    if model not in ["gpt-5-chat-latest"]:
        tools.append(
            WebSearchToolParam(
                type="web_search",
                search_context_size="high",
            )
        )
    if re.match(r"^o\d", model) or (re.match(r"gpt-5.*", model) and model != "gpt-5-chat-latest"):
        temperature = 1
    if re.match(r"gpt-5.*", model) and model != "gpt-5-chat-latest":
        reasoning = Reasoning(effort="high")
        tools.append(
            ImageGeneration(
                type="image_generation",
            )
        )

    if use_responses_api:
        if stream:
            return StreamOpenAIResponsesClient(
                model=model,
                messages=openai_messages,
                temperature=temperature,
                api_type=api_type,
                client=client,
                tools=tools,
                reasoning=reasoning,
            )
        else:
            return NonStreamOpenAIResponsesClient(
                model=model,
                messages=openai_messages,
                temperature=temperature,
                api_type=api_type,
                client=client,
                tools=tools,
                reasoning=reasoning,
            )
    else:
        if stream:
            return StreamOpenAIClient(
                model=model,
                messages=openai_messages,
                temperature=temperature,
                api_type=api_type,
                client=client,
                tools=tools,
                reasoning=reasoning,
            )
        else:
            return NonStreamOpenAIClient(
                model=model,
                messages=openai_messages,
                temperature=temperature,
                api_type=api_type,
                client=client,
                tools=tools,
                reasoning=reasoning,
            )



import tiktoken
from llm_bridge.type.chat_response import ChatResponse
from llm_bridge.type.model_message.openai_message import OpenAIMessage

from llm_bridge.type.model_message.openai_responses_message import OpenAIResponsesMessage


def count_openai_input_tokens(messages: list[OpenAIMessage]) -> int:
    text = ''
    file_count = 0

    for message in messages:
        for content in message.content:
            if content['type'] == "text":
                text += content['text']
            elif content['type'] in ("image_url", "input_audio"):
                file_count += 1

    return num_tokens_from_text(text) + file_count * 1000


def count_openai_responses_input_tokens(messages: list[OpenAIResponsesMessage]) -> int:
    text = ''
    file_count = 0

    for message in messages:
        for content in message['content']:
            if content['type'] in ("output_text", "input_text"):
                text += content['text']
            elif content['type'] in ("input_image", "input_file"):
                file_count += 1

    return num_tokens_from_text(text) + file_count * 1000


def count_openai_output_tokens(chat_response: ChatResponse) -> int:
    text = chat_response.text
    file_count = len(chat_response.files)

    return num_tokens_from_text(text) + file_count * 1000


def num_tokens_from_text(text: str) -> int:
    encoding = tiktoken.get_encoding("cl100k_base")
    return len(encoding.encode(text))



import base64

from llm_bridge.logic.file_fetch import fetch_file_data


async def get_raw_content_from_url(req_url: str) -> tuple[bytes, str]:
    file_data, media_type = await fetch_file_data(req_url)
    return file_data, media_type


# Base64 Encoded
async def get_encoded_content_from_url(req_url: str) -> tuple[str, str]:
    media_data, media_type = await get_raw_content_from_url(req_url)
    base64_media = base64.b64encode(media_data).decode('utf-8')
    return base64_media, media_type


async def get_openai_image_content_from_url(req_img_url: str) -> str:
    base64_image, media_type = await get_encoded_content_from_url(req_img_url)
    return f"data:{media_type};base64,{base64_image}"



import logging
from io import BytesIO

import fitz
import openpyxl
from docxlatex import Document as DocxLatexDocument
from fastapi import HTTPException
from pptx import Presentation

from llm_bridge.logic.file_fetch import fetch_file_data
from llm_bridge.logic.message_preprocess import file_type_checker


async def extract_text_from_file(file_url: str) -> str:
    file_content, content_type = await fetch_file_data(file_url)

    if file_type_checker.is_file_type_supported(file_url) is False:
        raise HTTPException(status_code=415,
                            detail=f"legacy filetypes ('.doc', '.xls', '.ppt') are not supported - {file_url}")

    file_type, sub_type = await file_type_checker.get_file_type(file_url)

    try:
        if sub_type == "code":
            return extract_text_from_code(file_content)
        if sub_type == "pdf":
            return extract_text_from_pdf(file_content)
        if sub_type == "word":
            return extract_text_from_word(file_content)
        if sub_type == "excel":
            return extract_text_from_excel(file_content)
        if sub_type == "powerpoint":
            return extract_text_from_ppt(file_content)
    except Exception as e:
        logging.exception(e)
        raise HTTPException(
            status_code=415,
            detail=f"Error during text extraction - {file_url} ({file_type}/{sub_type}) - {str(e)}"
        )

    try:
        return file_content.decode('utf-8')
    except UnicodeDecodeError as e:
        message = f"Unicode Decode Error during text extraction - {file_url} ({file_type}/{sub_type}) - {str(e)}"
        logging.error(message)
        return message


def extract_text_from_code(file_content: bytes) -> str:
    return file_content.decode('utf-8')


def extract_text_from_pdf(file_content: bytes) -> str:
    text = ""
    with fitz.open(stream=file_content, filetype="pdf") as doc:
        for page in doc:
            text += page.get_text()
    return text


def extract_text_from_word(file_content: bytes) -> str:
    with BytesIO(file_content) as f:
        doc = DocxLatexDocument(f)
        text_with_equations = doc.get_text()
        return text_with_equations


def extract_text_from_excel(file_content: bytes) -> str:
    text = ""
    workbook = openpyxl.load_workbook(BytesIO(file_content), data_only=True)
    for sheet in workbook:
        text += f"Sheet: {sheet.title}\n"
        for row in sheet.iter_rows(values_only=True):
            text += "\t".join([str(cell) if cell is not None else "" for cell in row]) + "\n"
    return text


def extract_text_from_ppt(file_content: bytes) -> str:
    text = ""
    presentation = Presentation(BytesIO(file_content))
    for slide in presentation.slides:
        for shape in slide.shapes:
            if hasattr(shape, "text"):
                text += shape.text + "\n"
    return text



import logging
import re
from pprint import pprint
from typing import Optional

import httpx
import openai
from fastapi import HTTPException
from openai import APIStatusError
from openai.types.responses import WebSearchToolParam, Response

from llm_bridge.client.implementations.openai.openai_token_couter import count_openai_responses_input_tokens, \
    count_openai_output_tokens
from llm_bridge.client.model_client.openai_client import OpenAIClient
from llm_bridge.type.chat_response import ChatResponse, Citation, File
from llm_bridge.type.serializer import serialize


def process_openai_responses_non_stream_response(
        response: Response,
        input_tokens: int,
) -> ChatResponse:

    output_list = response.output

    text: str = ""
    files: list[File] = []
    citations: list[Citation] = []

    for output in output_list:
        if output.type == "message":
            for content in output.content:
                if content.type == "output_text":
                    text += content.text
                # Citation is unavailable in OpenAI Responses API
                # if annotations := content.annotations:
                #     for annotation in annotations:
                #         citations.append(
                #             Citation(
                #                 text=content.text[annotation.start_index:annotation.end_index],
                #                 url=annotation.url
                #             )
                #         )
        # Image Generation untestable due to organization verification requirement
        # if output.type == "image_generation_call":
        #     file = File(
        #         name="generated_image.png",
        #         data=output.result,
        #         type="image/png",
        #     )
        #     files.append(file)

    chat_response = ChatResponse(text=text, files=files)
    output_tokens = count_openai_output_tokens(chat_response)
    return ChatResponse(
        text=text,
        files=files,
        citations=citations,
        input_tokens=input_tokens,
        output_tokens=output_tokens,
    )


class NonStreamOpenAIResponsesClient(OpenAIClient):
    async def generate_non_stream_response(self) -> ChatResponse:
        try:
            logging.info(f"messages: {self.messages}")

            input_tokens = count_openai_responses_input_tokens(
                messages=self.messages
            )

            response: Response = await self.client.responses.create(
                model=self.model,
                reasoning=self.reasoning,
                input=serialize(self.messages),
                temperature=self.temperature,
                stream=False,
                tools=self.tools,
            )

            return process_openai_responses_non_stream_response(
                response=response,
                input_tokens=input_tokens,
            )
        except httpx.HTTPStatusError as e:
            status_code = e.response.status_code
            text = e.response.text
            raise HTTPException(status_code=status_code, detail=text)
        except openai.BadRequestError as e:
            status_code = e.status_code
            text = e.message
            raise HTTPException(status_code=status_code, detail=text)
        except APIStatusError as e:
            status_code = e.status_code
            text = e.message
            raise HTTPException(status_code=status_code, detail=text)
        except Exception as e:
            logging.exception(e)
            match = re.search(r'\d{3}', str(e))
            if match:
                error_code = int(match.group(0))
            else:
                error_code = 500

            raise HTTPException(status_code=error_code, detail=str(e))


from dataclasses import dataclass
from typing import Optional


@dataclass
class File:
    name: str
    data: str
    type: str


@dataclass
class Citation:
    text: str
    indices: list[int]


# TODO: adapt to different Citation formats
# @dataclass
# class Citation:
#     text: str
#     indices: Optional[list[int]] = None
#     url: Optional[str] = None


@dataclass
class ChatResponse:
    text: Optional[str] = None
    thought: Optional[str] = None
    code: Optional[str] = None
    code_output: Optional[str] = None
    files: Optional[list[File]] = None
    display: Optional[str] = None
    citations: Optional[list[Citation]] = None
    error: Optional[str] = None
    input_tokens: Optional[int] = 0
    output_tokens: Optional[int] = 0
